/******************  dosmembk.c  ********************************************/

/*
   ***************************************************************
   ** This software may be used in both private and commercial  **
   ** environments, however it may not be sold for profit.      **
   **                                                           **
   **        Mike Miller, Box 105, Bethel, Conn. 06801          **
   ***************************************************************


   version 1.1 09apr87 - fixed a bug in getnxtmb().  If a memory
                        block was greater than 64K, unpredictable
                        results did occur.  Changed the constant 16
                        to 16L, as marked in the source below.



*/


#include <string.h>
#include <dos.h>
#include "dosmembk.h"

#ifndef I8086L
Must compile using the large memory model.
#endif






/******************************************************************************
 * Change the value of a variable in the DOS environment area.  The change is
 * made to the environment area that is associated with the original command
 * processor, not to copy of the environment area that the current program
 * is using.  Because of this, it is possible for a program to permanently
 * change the value of an environment variable.
 *
 * Returns:  0 if everything went as planned.
 *           1 if an error occurs.
 ***/
int setevar (pevar, pval)
   byte *pevar;         /* variable to change */
   byte *pval;          /* new value */
{
   membkhdr *pmbh;
   byte *pc;
   static byte envvar[128];     /* no special reason for 128 */
   int rc;


   if (strlen (pevar) > sizeof (envvar) - 1) {
      cprintf ("Maximum length of environment variable name is %d.\n",
                   sizeof (envvar));
      return (1);
      } /* if (strlen... */

   strcpy (envvar, pevar);      /* move to a known work area */
   strupr (envvar);             /* DOS capitalizes environ vars */
   strcat (envvar, "=");        /* search for var and equals sign */

   rc = get1stmb (&pmbh);       /* get first DOS memory block */
   if (rc == -1) {
      cprintf ("Unexpected error from get1stmb.\n");
      return (1);
      } /* if (rc == -1) */


   /* look through the first memory block to find the variable */
   /*  in question.  Usually it isn't here, but look anyway. */
   /*    (possible bug: if pmbh->bklen * 16 is > 32k)  */

   pc = varfind (pmbh->memblk, pmbh->bklen * 16, envvar, strlen (envvar));

   if (pc != NULL) {                    /* found the variable */
      cprintf (" Variable: %s\n", pc);

      rc = varchg (pc, pval);           /* change the value */
      if (rc != 0) {
         cprintf ("Unable to perform change.\n");
         return (1);
         } /* if (rc != 0) */

       else return (0);                 /* looks good */

      } /* if (pc != NULL) */


   /*  The variable wasn't in the first block (no surprise), so */
   /*  look through subsequent DOS memory blocks trying to find */
   /*  it.  */

   while (pmbh->em == 'M') {            /* last block will be 'Z' */

      rc = getnxtmb (&pmbh);            /* get next memory block */
      if (rc == -1) {
         cprintf ("Error getting next memory block.\n");
         return (1);
         } /* if (rc == -1) */

      /* look through each memory block to find the variable */
      /*  in question.                                       */
      /*    (possible bug: if pmbh->bklen * 16 is > 32k)     */

      pc = varfind (pmbh->memblk, pmbh->bklen * 16, envvar, strlen (envvar));

      if (pc != NULL) {                 /* found the variable */
         cprintf (" Variable: %s\n", pc);

         rc = varchg (pc, pval);        /* change its value */
         if (rc != 0) {
            cprintf ("Unable to perform change.\n");
            return (1);
            } /* if (rc != 0) */

          else return (0);              /* fini */

         } /* if (pc != NULL) */

      } /* while (pmbh->em == 'M') */

   cprintf ("Variable not found.\n");
   return (1);

   } /* int setevar (pevar, pval) */






/******************************************************************************
 * Look through a memory block for the variable specified.  Avoid false
 * recognitions based upon substrings (looking for ZAP= and actually finding
 * ABCZAP=).  This is accomplished by the following algorithm:  the first
 * byte of the variable found is either the first byte of the block, or the
 * byte before the first byte of the variable is a '\0'.
 *
 * Returns:   NULL if variable is not found
 *            pointer to var if it is found
 ***/
byte *varfind (blk, blklen, var, varlen)
   byte *blk;                   /* memory block */
   unsigned int blklen;         /* length of *blk */
   byte *var;                   /* variable to find */
   unsigned int varlen;         /* length of *var */
{
   byte *pv;                    /* ptr to variable found */
   byte *pb;                    /* working ptr to memory block */
   byte *pend;                  /* ptr to end of the block */

   extern byte *memfind (byte *, unsigned int, byte *, unsigned int);

   pend = blk + blklen;         /* for future reference */

   pb = blk;                    /* start at the beginning */
   while (1) {                  /* loop until found */

      pv = memfind (pb, blklen, var, varlen);   /* look for var string */
      if (pv == NULL) break; /* while (1) */    /* not found */

      /* found a matching string, see if it is the variable we want */

      if (pv == blk || *(pv - 1) == '\0')       /* looks good */
         break; /* while (1) */

      /* not the one we want, adjust pointers and lengths and look some more */

      pb = pv + 1;                              /* start at next byte */
      if (pb + varlen > pend) {                 /* at the end yet? */
         pv = NULL;                             /*  yes */
         break; /* while (1) */
         } /* if (pb > pend */

      blklen = blklen - (pb - blk);             /* new length */

      } /* while (1) */

   return (pv);

   } /* byte *varfind (blk, blklen, var, varlen) */







/******************************************************************************
 * Change an the value of the environment variable to the value specified.
 * Check to see if the new value is the same length as the old value, don't
 * make the change if the lengths are different (although I suppose if the
 * new value is smaller than the old one, it shouldn't matter.  But be
 * careful to avoid two '\0's in a row, DOS uses that to mark the end of
 * the environment area).
 *
 * Returns: 0 if OK
 *          1 if an error
 ***/
int varchg (pvar, pnewval)
   byte *pvar;          /* variable to change */
   byte *pnewval;       /* new value */
{
   byte *pc;                           /* working char pointer */


   pc = strchr (pvar, '=');             /* equals should always be there */
   if (pc == NULL) return (1);

   pc++;                                /* point to value of variable */
   if (strlen (pnewval) != strlen (pc)) {
      cprintf ("Length of the new value must");
      cprintf (" equal that of the old value.\n");
      return (1);
      } /* if (strlen... */

   strcpy (pc, pnewval);                /* change the value */

   cprintf ("New value: %s\n", pvar);   /* show what was done */

   return (0);

   } /* int varchg (pvar, pnewval) */







/******************************************************************************
 * Get the address of the header of the first DOS memory block and place the
 * address in the slot provided.
 *
 * Returns:  0 if the new memory block is the last memory block in the chain.
 *           1 if more memory blocks are left in the chain.
 *          -1 if something odd occurs.
 ***/
int get1stmb (ppmbhdr)
   membkhdr **ppmbhdr;  /* slot for memory block pointer.  CAUTION: it is */
                        /*  the address of the pointer.  */
{
   union REGSS reg;     /* register structure for interrupt */
   word *mbseg;
   word segaddr;
   int rc;


   /* get the first memory block that DOS allocated */
   /* (DOS function 0x52 is an undocumented function) */
   reg.x.ax = 0x5200;
   intdoss (&reg, &reg);

      /* ES:BX points to a memory location two bytes after the   */
      /*  word containing the segment of the first memory block, */
      /*  so subtract 2.                                         */
      /*                ES:BX ===|                               */
      /*                         v                               */
      /*       | mbseg | byte | byte |                           */

   mbseg = (word *) ((16 * reg.x.es) + (reg.x.bx - 2));

   segaddr = *mbseg;

      /* segaddr now contains the segment of the first memory    */
      /*  block.  Convert to a full address, and place into the  */
      /*  slot provided.                                         */

   *ppmbhdr = (membkhdr *)(segaddr * 16);

      /* *ppmbhdr now contains the address of the first memory   */
      /*  block. Check to see if there are any blocks past the   */
      /*  first one and set the return code accordingly.         */

   if ((*ppmbhdr)->em == 'M') rc = 1;           /* more blocks */
    else if ((*ppmbhdr)->em == 'Z') rc = 0;     /* last block */
    else rc = -1;                               /* unknown */

   return (rc);

   } /* int get1stmb (ppmbhdr) */






/******************************************************************************
 * Get the address of the header of the next DOS memory block in the chain,
 * given the address of the current block.
 *
 * Returns:  0 if the new memory block is the last memory block in the chain.
 *           1 if more memory blocks are left in the chain.
 *          -1 if something odd occurs.
 ***/
int getnxtmb (ppmbhdr)
   membkhdr **ppmbhdr;  /* same as get1stmb */
{
   int rc;
   byte *byteptr;

   byteptr = (*ppmbhdr)->memblk;        /* start at the beginning of the */
                                        /*  current memory block */

                         /* version 1.1 */
                         /*           v */
   byteptr += ((*ppmbhdr)->bklen) * 16L; /* bklen is in paragraphs */

   *ppmbhdr = (membkhdr *)byteptr;      /* put back into slot */

   if ((*ppmbhdr)->em == 'M') rc = 1;           /* more blocks */
    else if ((*ppmbhdr)->em == 'Z') rc = 0;     /* last block */
    else rc = -1;                               /* unknown */

   return (rc);

   } /* int getnxtmb (ppmbhdr) */
